from typing import Tuple, FrozenSet, Optional


def all_words() -> Tuple[str, ...]:
    pass


def codepoints_for_word(word: str) -> FrozenSet[int]:
    pass


def name_for_codepoint(cp: int) -> Optional[str]:
    pass
